package activitypub

  1. Overview
  2. Docs

Class type Types.object_Source

An Object's interface.

Most of the methods are accessors, querying the internal #g graph. Before querying any information, it is important that the object have been dererefenced using the #dereference method. Calling #dereference on an already dereferenced object has no effect, i.e. the internal graph is not updated.

method id : id
method as_id : Iri.t option
method iri : Iri.t
method type_ : Iri.t
method attachment : [ `L of link | `O of object_ ] list
method attributed_to : [ `L of link | `O of object_ ] option
method audience : [ `L of link | `O of object_ ] option
method likes : collection option
method shares : collection option
method content : string option
method content_map : string Activitypub.Smap.t
method name : string option
method name_map : string Activitypub.Smap.t
method end_time : Rdf.Term.datetime option
method generator : [ `L of link | `O of object_ ] option
method icon : [ `L of link | `I of image ] list
method image : [ `L of link | `I of image ] list
method in_reply_to : [ `L of link | `O of object_ ] list
method preview : [ `L of link | `O of object_ ] option
method published : Rdf.Term.datetime option
method replies : collection option
method start_time : Rdf.Term.datetime option
method summary : string option
method summary_map : string Activitypub.Smap.t
method tag : [ `L of link | `O of object_ ] list
method updated : Rdf.Term.datetime option
method url : [ `L of link | `I of Iri.t ] list
method to_ : [ `L of link | `O of object_ ] list
method bto : [ `L of link | `O of object_ ] list
method cc : [ `L of link | `O of object_ ] list
method bcc : [ `L of link | `O of object_ ] list
method media_type : Ldp.Ct.mime option
method g : Rdf.Graph.graph option
method is_empty : bool
method pp : Format.formatter -> unit -> unit
method dereference : unit Lwt.t
method as_activity : activity