package activitypub

  1. Overview
  2. Docs

Source file e.ml

1
2
3
4
5
6
7
8
9
10
11
12
13
14
15
16
17
18
19
20
21
22
23
24
25
26
27
28
29
30
31
32
33
34
35
36
37
38
39
40
41
42
43
44
45
46
47
48
49
50
51
52
(*********************************************************************************)
(*                OCaml-ActivityPub                                              *)
(*                                                                               *)
(*    Copyright (C) 2023-2024 INRIA All rights reserved.                         *)
(*    Author: Maxence Guesdon, INRIA Saclay                                      *)
(*                                                                               *)
(*    This program is free software; you can redistribute it and/or modify       *)
(*    it under the terms of the GNU Lesser General Public License version        *)
(*    3 as published by the Free Software Foundation.                            *)
(*                                                                               *)
(*    This program is distributed in the hope that it will be useful,            *)
(*    but WITHOUT ANY WARRANTY; without even the implied warranty of             *)
(*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              *)
(*    GNU General Public License for more details.                               *)
(*                                                                               *)
(*    You should have received a copy of the GNU General Public License          *)
(*    along with this program; if not, write to the Free Software                *)
(*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA                   *)
(*    02111-1307  USA                                                            *)
(*                                                                               *)
(*    Contact: maxence.guesdon@inria.fr                                          *)
(*                                                                               *)
(*********************************************************************************)

(** Errors. *)

(** Extensible error type. When adding a new constructor,
 do not forget to register a printer with {!register_string_of_error}. *)
type error = ..

exception Error of error
let error e = raise (Error e)

let string_of_error_funs = ref []
let register_string_of_error f =
  string_of_error_funs := f :: !string_of_error_funs
let string_of_error =
  let rec iter e = function
  | [] -> "No printer for error"
  | f :: q ->
      match f e with
      | Some s -> s
      | None -> iter e q
  in
  fun e -> iter e !string_of_error_funs

let pp ppf e = Format.fprintf ppf "%s" (string_of_error e)

let () = Printexc.register_printer
  (function Error e -> Some (string_of_error e) | _ -> None)